﻿using MySql.Data.MySqlClient;
using Org.BouncyCastle.Asn1.X509.Qualified;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// The outline of a MySqlColumn attribute that specify's information about a field
    /// and how it's setup in a MySql Row.
    /// </summary>
    [AttributeUsage(AttributeTargets.Field, AllowMultiple = false, Inherited = true)]
    public class MySqlColumn : Attribute
    {
        /// <summary>
        /// Whether or not this column is the row id.
        /// </summary>
        public bool Id = false;

        /// <summary>
        /// The name of the column, if null, the field name will be used instead.
        /// </summary>
        public string Name = null;

        /// <summary>
        /// The MySql Data type of the column, if null, the field type will be used instead.
        /// </summary>
        public string Type = null;

        /// <summary>
        /// Whether or not this column is a primary key.
        /// </summary>
        public bool PrimaryKey = false;

        /// <summary>
        /// Whether or not this column auto increments.
        /// </summary>
        public bool AutoIncrement = false;

        /// <summary>
        /// An overrided default value for this column if set, default is null.
        /// </summary>
        public object DefaultValue = null;

        /// <summary>
        /// An optional MySqlColumnConverter that can convert this column to another
        /// data type when needed.
        /// </summary>
        public Type Converter = null;

        /// <summary>
        /// Creates a new default MySqlColumn.
        /// </summary>
        public MySqlColumn() { }

        /// <summary>
        /// Creates a new MySqlColumn with a name and column converter if needed.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="converter"></param>
        /// <exception cref="NotSupportedException"></exception>
        public MySqlColumn(string name, Type converter = null)
        {
            this.Name = name;

            //Make sure the converter is valid if one was passed.
            if (converter != null && !converter.IsAssignableTo(typeof(MySqlColumnConverter)))
                throw new NotSupportedException($"Converter must inherit {nameof(MySqlColumnConverter)}");

            this.Converter = converter;
        }

        /// <summary>
        /// Creates a new MySqlColumn with a name and type and optional column converter.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="type"></param>
        /// <param name="converter"></param>
        public MySqlColumn(string name, string type, Type converter = null)
        {
            this.Name = name;

            this.Type = type;

            //Make sure the converter is valid if one was passed.
            if (converter != null && !converter.IsAssignableTo(typeof(MySqlColumnConverter)))
                throw new NotSupportedException($"Converter must inherit {nameof(MySqlColumnConverter)}");

            this.Converter = converter;
        }
    }
}
