﻿using MySql.Data.MySqlClient;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// A set of extensions to help work with MySqlCommands.
    /// </summary>
    public static class MySqlCommandExtensions
    {
        /// <summary>
        /// Setups this MySqlCommand to insert a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Insert<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the insert section.
            builder.Append($"INSERT INTO `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Write the set values section.
            builder.Append("SET ");

            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //Skip id columns because they are auto incremented.
                if (column.Id)
                    continue;

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", (column.Converter == null ? fields[i].GetValue(row) : column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(fields[i].GetValue(row))));

                seperate = true;
            }

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to update a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Update<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the update section
            builder.Append($"UPDATE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Write the set values section
            builder.Append("SET ");

            FieldInfo idField = null;
            MySqlColumn idColumn = null;
            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //Skip id columns because they are auto incremented.
                if (column.Id)
                {
                    idField = fields[i];
                    idColumn = column;
                    continue;
                }

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", (column.Converter == null ? fields[i].GetValue(row) : column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(fields[i].GetValue(row))));

                seperate = true;
            }

            //Make sure we have an id column.
            if (idField == null)
                throw new Exception("Given row does not contain an id column");

            //Write the where clause
            builder.Append($" WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            command.Parameters.AddWithValue("@id", idField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to get a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        /// <exception cref="Exception"></exception>
        public static void Get<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get the id field
            if (!rowType.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find the id column on the given row.");

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1";
            
            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to get all the rows of a given type from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void GetAll<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`";
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id from an instance of the row.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Delete<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Get the id the column and field info
            if (!row.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find Id column on row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", idField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Delete<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Get the id the column and field info
            if (!rowType.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find Id column on row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete all rows of a given type from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void DeleteAll<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to create a table from a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void CreateTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Create a new instance of the row
            var row = rowType.CreateInstance<T>();

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"CREATE TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` (");

            //Write all the columns
            var fields = rowType.GetFields();

            bool separate = false;

            if (fields != null)
            {
                for (int i = 0; i < fields.Length; i++)
                {
                    var field = fields[i];

                    var column = field.GetCustomAttribute<MySqlColumn>();

                    if (column != null)
                    {
                        if (separate)
                            builder.Append(", ");

                        //Write the column name
                        builder.Append($"{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)} ");

                        //Write the column data type
                        builder.Append($"{column.GetMySqlColumnType(fields[i])} ");

                        //Determine if this column is nullable
                        bool nullable = column.IsNullable(field, row);

                        //Write the column null information
                        builder.Append($"{(nullable ? "NULL" : "NOT NULL")} ");

                        //Write the column default value if needed. (We can't do this if the column is auto increment)
                        if (!column.AutoIncrement)
                        {
                            var defaultValue = column.DefaultValue;

                            //If we don't have a default value, attempt to get it from the field.
                            if (defaultValue == null)
                                defaultValue = field.GetValue(row);

                            //If we have a converter, run the converter to get the true default value.
                            if (column.Converter != null)
                                defaultValue = column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(defaultValue);

                            //If we have a default value, set it on the column.
                            if (defaultValue != null)
                            {
                                builder.Append($"DEFAULT @{i} ");

                                command.Parameters.AddWithValue($"@{i}", defaultValue);
                            }
                        }

                        //Write the column auto increment information.
                        builder.Append($"{(column.AutoIncrement ? "AUTO_INCREMENT" : "")}");

                        if (column.PrimaryKey)
                            builder.Append($", PRIMARY KEY ({(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)})");

                        separate = true;
                    }
                }
            }

            //Add any indexes
            var indexes = rowType.GetCustomAttributes<MySqlRowIndex>().ToList();
            if (indexes != null)
            {
                for (int i = 0; i < indexes.Count; i++)
                {
                    var index = indexes[i];

                    if (index.Columns == null || index.Columns.Length == 0)
                        continue;

                    if (separate)
                        builder.Append(", ");

                    builder.Append($"INDEX {index.Name} ({index.Columns.Aggregate((curr, next) => curr == null ? next : curr + ", " + next)})");

                    separate = true;
                }
            }

            builder.Append(")");

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to check if a table exists for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void TableExists<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"SHOW TABLES LIKE '{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}'");

            //Set the command text
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to empty the table for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void EmptyTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"TRUNCATE TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a table for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void DeleteTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DROP TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text
            command.CommandText = builder.ToString();
        }
    }
}
