﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// A set of extensions to help work with Types.
    /// </summary>
    internal static class TypeExentions
    {
        /// <summary>
        /// Creates an instance of a given type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="type"></param>
        /// <returns></returns>
        public static T CreateInstance<T>(this Type type)
        {
            if (type == null)
                return default;

            //See if we can create an instance of the type the easy way.
            try { return (T)Activator.CreateInstance(type); }
            catch { }

            //See if we can force create the type..
            try { return (T)System.Runtime.Serialization.FormatterServices.GetUninitializedObject(type); }
            catch { }

            //If that failed, then this might be an array, so try again.
            try { return (T)Activator.CreateInstance(type, 0); }
            catch { }

            //If we get here it means we couldn't create an instance so return null.
            return default;
        }

        /// <summary>
        /// Converts a given object to a given type with better auto conversion support.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="obj"></param>
        /// <returns></returns>
        /// <exception cref="NotSupportedException"></exception>
        public static T ConvertToType<T>(this object obj)
        {
            if (obj == null)
                return default(T);

            var target = Type.GetTypeCode(typeof(T));

            var current = Type.GetTypeCode(obj.GetType());

            if (target == current)
            {
                return (T)obj;
            }
            else
            {
                switch (target)
                {
                    case TypeCode.Boolean:
                        switch (current)
                        {
                            case TypeCode.Byte:
                                return (T)(object)((byte)obj != 0 ? true : false);
                            case TypeCode.Char:
                                return (T)(object)((char)obj != 'f' && (char)obj != 'F' && (char)obj != '0' ? true : false);
                            case TypeCode.Decimal:
                                return (T)(object)((decimal)obj > 0 ? true : false);
                            case TypeCode.Double:
                                return (T)(object)((double)obj > 0 ? true : false);
                            case TypeCode.Int16:
                                return (T)(object)((short)obj > 0 ? true : false);
                            case TypeCode.Int32:
                                return (T)(object)((int)obj > 0 ? true : false);
                            case TypeCode.Int64:
                                return (T)(object)((long)obj > 0 ? true : false);
                            case TypeCode.SByte:
                                return (T)(object)((sbyte)obj > 0 ? true : false);
                            case TypeCode.Single:
                                return (T)(object)((float)obj > 0 ? true : false);
                            case TypeCode.String:
                                {
                                    var str = ((string)obj).ToLower().Trim();

                                    if (str == "t" || str == "1" || str == "true" || str == "yes" || str == "y")
                                        return (T)(object)true;
                                    else
                                        return (T)(object)false;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)((ushort)obj > 0 ? true : false);
                            case TypeCode.UInt32:
                                return (T)(object)((uint)obj > 0 ? true : false);
                            case TypeCode.UInt64:
                                return (T)(object)((ulong)obj > 0 ? true : false);
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)false;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Byte:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Char:
                                return (T)(object)(byte)(char)obj;
                            case TypeCode.Decimal:
                                return (T)(object)(byte)(decimal)obj;
                            case TypeCode.Double:
                                return (T)(object)(byte)(double)obj;
                            case TypeCode.Int16:
                                return (T)(object)(byte)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(byte)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(byte)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(byte)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(byte)(float)obj;
                            case TypeCode.String:
                                {
                                    if (byte.TryParse((string)obj, out byte result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(byte)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(byte)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(byte)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Char:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? '1' : '0');
                            case TypeCode.Byte:
                                return (T)(object)(char)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(char)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(char)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(char)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(char)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(char)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(char)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(char)(double)obj;
                            case TypeCode.String:
                                return (T)(object)((string)obj)[0];
                            case TypeCode.UInt16:
                                return (T)(object)(char)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(char)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(char)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return default(T);
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.DateTime:
                        switch (current)
                        {
                            case TypeCode.String:
                                if (DateTime.TryParse((string)obj, out DateTime result))
                                    return (T)(object)result;
                                else
                                    return default(T);
                            default:
                                return default(T);
                        }
                    case TypeCode.DBNull:
                        return default(T);
                    case TypeCode.Decimal:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(decimal)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(decimal)(char)obj;
                            case TypeCode.Double:
                                return (T)(object)(decimal)(double)obj;
                            case TypeCode.Int16:
                                return (T)(object)(decimal)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(decimal)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(decimal)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(decimal)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(decimal)(float)obj;
                            case TypeCode.String:
                                {
                                    if (decimal.TryParse((string)obj, out decimal result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(decimal)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(decimal)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(decimal)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Double:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(double)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(double)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(double)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(double)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(double)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(double)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(double)(float)obj;
                            case TypeCode.String:
                                {
                                    if (double.TryParse((string)obj, out double result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(double)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(double)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(double)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Empty:
                        return default(T);
                    case TypeCode.Int16:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(short)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(short)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(short)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(short)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(short)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(short)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(short)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(short)(double)obj;
                            case TypeCode.String:
                                {
                                    if (short.TryParse((string)obj, out short result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(short)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(short)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(short)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Int32:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(int)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(int)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(int)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(int)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(int)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(int)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(int)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(int)(double)obj;
                            case TypeCode.String:
                                {
                                    if (int.TryParse((string)obj, out int result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(int)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(int)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(int)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Int64:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(long)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(long)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(long)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(long)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(long)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(long)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(long)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(long)(double)obj;
                            case TypeCode.String:
                                {
                                    if (long.TryParse((string)obj, out long result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(long)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(long)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(long)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Object:
                        return (T)obj;
                    case TypeCode.SByte:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(sbyte)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(sbyte)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(sbyte)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(sbyte)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(sbyte)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(sbyte)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(sbyte)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(sbyte)(double)obj;
                            case TypeCode.String:
                                {
                                    if (sbyte.TryParse((string)obj, out sbyte result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(sbyte)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(sbyte)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(sbyte)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.Single:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(float)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(float)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(float)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(float)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(float)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(float)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(float)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(float)(double)obj;
                            case TypeCode.String:
                                {
                                    if (float.TryParse((string)obj, out float result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(float)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(float)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(float)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.String:
                        switch (current)
                        {
                            case TypeCode.Empty:
                            case TypeCode.DBNull:
                                return default(T);
                            default:
                                return (T)(object)obj.ToString();
                        }
                    case TypeCode.UInt16:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(ushort)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(ushort)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(ushort)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(ushort)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(ushort)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(ushort)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(ushort)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(ushort)(double)obj;
                            case TypeCode.String:
                                {
                                    if (ushort.TryParse((string)obj, out ushort result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(ushort)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(ushort)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(ushort)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.UInt32:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(uint)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(uint)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(uint)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(uint)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(uint)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(uint)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(uint)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(uint)(double)obj;
                            case TypeCode.String:
                                {
                                    if (uint.TryParse((string)obj, out uint result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(uint)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(uint)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(uint)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    case TypeCode.UInt64:
                        switch (current)
                        {
                            case TypeCode.Boolean:
                                return (T)(object)((bool)obj == true ? 1 : 0);
                            case TypeCode.Byte:
                                return (T)(object)(ulong)(byte)obj;
                            case TypeCode.Char:
                                return (T)(object)(ulong)(char)obj;
                            case TypeCode.Int16:
                                return (T)(object)(ulong)(short)obj;
                            case TypeCode.Int32:
                                return (T)(object)(ulong)(int)obj;
                            case TypeCode.Int64:
                                return (T)(object)(ulong)(long)obj;
                            case TypeCode.SByte:
                                return (T)(object)(ulong)(sbyte)obj;
                            case TypeCode.Single:
                                return (T)(object)(ulong)(float)obj;
                            case TypeCode.Double:
                                return (T)(object)(ulong)(double)obj;
                            case TypeCode.String:
                                {
                                    if (ulong.TryParse((string)obj, out ulong result))
                                        return (T)(object)result;
                                    else
                                        return (T)(object)0;
                                }
                            case TypeCode.UInt16:
                                return (T)(object)(ulong)(ushort)obj;
                            case TypeCode.UInt32:
                                return (T)(object)(ulong)(uint)obj;
                            case TypeCode.UInt64:
                                return (T)(object)(ulong)(ulong)obj;
                            case TypeCode.Object:
                            case TypeCode.DBNull:
                            case TypeCode.Empty:
                            case TypeCode.DateTime:
                                return (T)(object)0;
                            default:
                                throw new NotSupportedException($"Unsupported current typecode: {current}");
                        }
                    default:
                        throw new NotSupportedException($"Unsupported target typecode: {target}");
                }
            }
        }

        /// <summary>
        /// Returns whether or not a given type is a System Nullabe type.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public static bool IsNullable(this Type type)
        {
            return type.IsGenericType && type.GetGenericTypeDefinition().IsEquivalentTo(typeof(Nullable<>));
        }
    }
}
